///////////////////////////////////////////////////////////////////////////////
//
// 3D Utility Library
// Copyright (C) 2006 Tobias Minch
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Written by
//   Tobias Minich - belgabor@gmx.de
//
// Parts of this code are based on code written for rct3tool (Copyright 2005
// Jonathan Wilson).
//
///////////////////////////////////////////////////////////////////////////////

#include "SCNFile.h"
#include "matrix.h"
#include "3DLoader.h"
#include <string>
#include <sstream>
#include <stdio.h>


bool cSCNFile::Load() {
    char *tmp;
    unsigned long i;
    error = CSCNFILE_NO_ERROR;

    if (filename == "") {
        error = CSCNFILE_ERROR_FILENAME;
        return false;
    }

    FILE *f = fopen(filename.c_str(), "rb");
    if (!f) {
        error = CSCNFILE_ERROR_FILENAME;
        return false;
    }

    unsigned long objlen;
    fread(&objlen, sizeof(objlen), 1, f);

    tmp = new char[MAX_PATH];
    if (objlen)
        fread(tmp, objlen, 1, f);

    if (objfile == "") {
        if (objlen)
            objfile = tmp;
    }
    delete[] tmp;

    unsigned long meshcount;
    fread(&meshcount, sizeof(meshcount), 1, f);
    meshstructs.clear();
    for (i = 0; i < meshcount; i++) {
        cMeshStruct ms;
        fread(&ms.disabled, sizeof(ms.disabled), 1, f);
        if (ms.disabled == false) {
            fread(&ms.flags, sizeof(ms.flags), 1, f);
            fread(&ms.place, sizeof(ms.place), 1, f);
            fread(&ms.unknown, sizeof(ms.unknown), 1, f);

            unsigned long ftxlen;
            fread(&ftxlen, sizeof(ftxlen), 1, f);
            if (ftxlen) {
                tmp = new char[ftxlen];
                fread(tmp, ftxlen, 1, f);
                ms.FTX = tmp;
                delete[] tmp;
            } else
                ms.FTX = "";

            unsigned long txslen;
            fread(&txslen, sizeof(txslen), 1, f);
            if (txslen) {
                tmp = new char[txslen];
                fread(tmp, txslen, 1, f);
                ms.TXS = tmp;
                delete[] tmp;
            } else
                ms.TXS = "";
        }
        ms.effectpoint = false;
        meshstructs.push_back(ms);
    }

    flexitextures.clear();
    unsigned long texcount;
    fread(&texcount, sizeof(texcount), 1, f);
    for (i = 0; i < texcount; i++) {
        cFlexiTexture ft;
        fread(&ft.Recolorable, sizeof(ft.Recolorable), 1, f);

        unsigned long namelen;
        fread(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            tmp = new char[namelen];
            fread(tmp, namelen, 1, f);
            ft.Name = tmp;
            delete[] tmp;
        } else
            ft.Name = "";

        unsigned long texturelen;
        fread(&texturelen, sizeof(texturelen), 1, f);
        if (texturelen) {
            tmp = new char[texturelen];
            fread(tmp, texturelen, 1, f);
            ft.Texture = tmp;
            delete[] tmp;
        } else
            ft.Texture = "";

        unsigned long alphalen;
        fread(&alphalen, sizeof(alphalen), 1, f);
        if (alphalen) {
            tmp = new char[alphalen];
            fread(tmp, alphalen, 1, f);
            ft.Alpha = tmp;
            delete[] tmp;
        } else
            ft.Alpha = "";

        flexitextures.push_back(ft);
    }

    references.clear();
    unsigned long refcount;
    fread(&refcount, sizeof(refcount), 1, f);
    for (i = 0; i < refcount; i++) {
        unsigned long reflen;
        fread(&reflen, sizeof(reflen), 1, f);
        if (reflen) {
            std::string ref;
            tmp = new char[reflen];
            fread(tmp, reflen, 1, f);
            ref = tmp;
            delete[] tmp;
            references.push_back(ref);
        }
    }

    effectpoints.clear();
    unsigned long ptcount;
    fread(&ptcount, sizeof(ptcount), 1, f);
    for (i = 0; i < ptcount; i++) {
        cEffectPoint e;

        unsigned long namelen;
        fread(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            tmp = new char[namelen];
            fread(tmp, namelen, 1, f);
            e.Name = tmp;
            delete[] tmp;
        } else
            e.Name = "";

        D3DMATRIX m;
        fread(&m._11, sizeof(m._11), 1, f);
        fread(&m._12, sizeof(m._12), 1, f);
        fread(&m._13, sizeof(m._13), 1, f);
        fread(&m._14, sizeof(m._14), 1, f);
        fread(&m._21, sizeof(m._21), 1, f);
        fread(&m._22, sizeof(m._22), 1, f);
        fread(&m._23, sizeof(m._23), 1, f);
        fread(&m._24, sizeof(m._24), 1, f);
        fread(&m._31, sizeof(m._31), 1, f);
        fread(&m._32, sizeof(m._32), 1, f);
        fread(&m._33, sizeof(m._33), 1, f);
        fread(&m._34, sizeof(m._34), 1, f);
        fread(&m._41, sizeof(m._41), 1, f);
        fread(&m._42, sizeof(m._42), 1, f);
        fread(&m._43, sizeof(m._43), 1, f);
        fread(&m._44, sizeof(m._44), 1, f);
        e.Transform.push_back(m);
        e.TransformNames.push_back("Custom Matrix (old scn)");

        effectpoints.push_back(e);
    }

    tmpmeshlist.clear();
    transforms.clear();
    transformnames.clear();

    if (!feof(f)) {
        version = 0;
        fread(&version, sizeof(long), 1, f);

        if (version >= 1) {
            long fixorient = 0;
            fread(&fixorient, sizeof(long), 1, f);
            if ((version <=2) && fixorient) {
                transforms.push_back(matrixGetLegacyFixOrientation());
                transformnames.push_back("Fix Orientation (old scn)");
            }
        }
        // SCN Version 2 Stuff
        if (version >= 2) {
            fread(&sivsettings.sivflags, sizeof(unsigned int), 1, f);
            fread(&sivsettings.sway, sizeof(float), 1, f);
            fread(&sivsettings.brightness, sizeof(float), 1, f);
            fread(&sivsettings.unknown, sizeof(float), 1, f);
            fread(&sivsettings.scale, sizeof(float), 1, f);
        }
        // SCN Version 3 Stuff
        if (version >= 3) {
            // Read Mesh names
            for (i = 0; i < meshcount; i++) {
                std::string mname;

                unsigned long namelen;
                fread(&namelen, sizeof(namelen), 1, f);
                if (namelen) {
                    tmp = new char[namelen];
                    fread(tmp, namelen, 1, f);
                    mname = tmp;
                    delete[] tmp;
                } else
                    mname = "";
                tmpmeshlist.push_back(mname);
            }

            // Read transformations
            unsigned long matrices;
            fread(&matrices, sizeof(matrices), 1, f);
            for (unsigned long j = 0; j < matrices; j++) {
                D3DMATRIX m;
                std::string nam;

                unsigned long namelen;
                fread(&namelen, sizeof(namelen), 1, f);
                if (namelen) {
                    tmp = new char[namelen];
                    fread(tmp, namelen, 1, f);
                    nam = tmp;
                    delete[] tmp;
                } else
                    nam = "";
                transformnames.push_back(nam);

                fread(&m._11, sizeof(m._11), 1, f);
                fread(&m._12, sizeof(m._12), 1, f);
                fread(&m._13, sizeof(m._13), 1, f);
                fread(&m._14, sizeof(m._14), 1, f);
                fread(&m._21, sizeof(m._21), 1, f);
                fread(&m._22, sizeof(m._22), 1, f);
                fread(&m._23, sizeof(m._23), 1, f);
                fread(&m._24, sizeof(m._24), 1, f);
                fread(&m._31, sizeof(m._31), 1, f);
                fread(&m._32, sizeof(m._32), 1, f);
                fread(&m._33, sizeof(m._33), 1, f);
                fread(&m._34, sizeof(m._34), 1, f);
                fread(&m._41, sizeof(m._41), 1, f);
                fread(&m._42, sizeof(m._42), 1, f);
                fread(&m._43, sizeof(m._43), 1, f);
                fread(&m._44, sizeof(m._44), 1, f);
                transforms.push_back(m);
            }

            // Read new effect point structures
            fread(&ptcount, sizeof(ptcount), 1, f);
            for (i = 0; i < ptcount; i++) {
                cEffectPoint e;

                unsigned long namelen;
                fread(&namelen, sizeof(namelen), 1, f);
                if (namelen) {
                    tmp = new char[namelen];
                    fread(tmp, namelen, 1, f);
                    e.Name = tmp;
                    delete[] tmp;
                } else
                    e.Name = "";

                fread(&matrices, sizeof(matrices), 1, f);
                for (unsigned long j = 0; j < matrices; j++) {
                    D3DMATRIX m;
                    std::string nam;

                    fread(&namelen, sizeof(namelen), 1, f);
                    if (namelen) {
                        tmp = new char[namelen];
                        fread(tmp, namelen, 1, f);
                        nam = tmp;
                        delete[] tmp;
                    } else
                        nam = "";
                    e.TransformNames.push_back(nam);

                    fread(&m._11, sizeof(m._11), 1, f);
                    fread(&m._12, sizeof(m._12), 1, f);
                    fread(&m._13, sizeof(m._13), 1, f);
                    fread(&m._14, sizeof(m._14), 1, f);
                    fread(&m._21, sizeof(m._21), 1, f);
                    fread(&m._22, sizeof(m._22), 1, f);
                    fread(&m._23, sizeof(m._23), 1, f);
                    fread(&m._24, sizeof(m._24), 1, f);
                    fread(&m._31, sizeof(m._31), 1, f);
                    fread(&m._32, sizeof(m._32), 1, f);
                    fread(&m._33, sizeof(m._33), 1, f);
                    fread(&m._34, sizeof(m._34), 1, f);
                    fread(&m._41, sizeof(m._41), 1, f);
                    fread(&m._42, sizeof(m._42), 1, f);
                    fread(&m._43, sizeof(m._43), 1, f);
                    fread(&m._44, sizeof(m._44), 1, f);
                    e.Transform.push_back(m);
                }

                effectpoints.push_back(e);
            }
        }
    }
    fclose(f);

    if (objfile == "")
        return true;

    return LoadObject(false);
}

cMeshStruct cSCNFile::MakeMeshStruct(c3DLoader *obj, unsigned int j) {
    cMeshStruct ms;
    if ((strcmp(obj->GetObjectName(j).c_str(), "root") == 0)
            || (obj->GetObjectVertexCount(j) < 3)
            || (!obj->IsObjectValid(j))) {
        ms.disabled = true;
    } else {
        ms.disabled = false;
    }
    if (obj->GetObjectVertexCount(j) == 1) {
        ms.effectpoint = true;
        ms.effectpoint_vert = obj->GetObjectVertex(j, 0).position;
    } else {
        ms.effectpoint = false;
    }
    ms.Name = obj->GetObjectName(j);
    ms.flags = 0;
    ms.place = 0;
    ms.unknown = 3;
    ms.FTX = "";
    ms.TXS = "";
    return ms;
}

void cSCNFile::ResetAndReload(c3DLoader *obj) {
    meshstructs.clear();
    // Just load the new object into new mesh structs
    for (long j = 0; j < obj->GetObjectCount(); j++) {
        meshstructs.push_back(MakeMeshStruct(obj, j));
    }

}

void cSCNFile::Fixup(c3DLoader *obj) {
    std::vector<cMeshStruct> fixup;
    int msinfile = meshstructs.size();
    unsigned int i;

    error = CSCNFILE_ERROR_MESH_FIX;

    // First round of fixups
    int nrfound = 0;
    for(i = 0; i < obj->GetObjectCount(); i++) {
        bool found = false;
        for(cMeshStructIterator ms = meshstructs.begin(); ms != meshstructs.end(); ms++) {
            if (ms->Name == obj->GetObjectName(i)) {
                // Found!
                fixup.push_back(*ms);
                found = true;
                nrfound++;
                meshstructs.erase(ms);
                break;
            }
        }
        if (!found)
            fixup.push_back(MakeMeshStruct(obj, i));
    }

    // If the fixups were successful, meshstructs is empty
    if (meshstructs.size() == 0) {
        if (msinfile != obj->GetObjectCount())
            error |= CSCNFILE_ERROR_MESH_MISMATCH;
        meshstructs = fixup;
        return;
    }

    // Didn't work. Yay! -.-
    if (obj->GetObjectCount() == meshstructs.size() == msinfile) {
        // Special case: Nothing worked, but we have the same number of meshes/objects
        error |= CSCNFILE_ERROR_MESH_CONFUSION;
        for(i = 0; i < obj->GetObjectCount(); i++) {
            meshstructs[i].Name = obj->GetObjectName(i);
        }
        return;
    }
    if (nrfound == obj->GetObjectCount()) {
        // Special case: Meshes/Objects disappeared
        error |= CSCNFILE_ERROR_MESH_POOF;
        meshstructs.clear();
        meshstructs = fixup;
        return;
    }

    // Well, we can't call this a fix, but it should be the best
    error = CSCNFILE_ERROR_MESH_CONFUSION;
    meshstructs.clear();
    meshstructs = fixup;
    return;
}

bool cSCNFile::LoadObject(bool reset) {
    unsigned int i;
    c3DLoader *obj = c3DLoader::LoadFile(const_cast<char *>(objfile.c_str()));
    if (reset)
        meshstructs.clear();

    if (!obj) {
        error = CSCNFILE_ERROR_OBJNOTFOUND;
        if (reset)
            return false;
        if (tmpmeshlist.size()) {
            for (i = 0; i < meshstructs.size(); i++)
                meshstructs[i].Name = tmpmeshlist[i];
        } else {
            for (i = 0; i < meshstructs.size(); i++) {
                std::stringstream fmt;
                fmt << "Mesh #" << (i+1);
                meshstructs[i].Name = fmt.str();
            }
        }
        return true;
    }

    if (reset) {
        ResetAndReload(obj);
    } else {
        if (obj->GetObjectCount() == meshstructs.size()) {
            if (tmpmeshlist.size()) {
                bool needsfix = false;
                for (i = 0; i < meshstructs.size(); i++) {
                    if (tmpmeshlist[i] != obj->GetObjectName(i))
                        needsfix = true;
                    meshstructs[i].Name = tmpmeshlist[i];
                    if (obj->GetObjectVertexCount(i) == 1) {
                        meshstructs[i].effectpoint = true;
                        meshstructs[i].effectpoint_vert = obj->GetObjectVertex(i, 0).position;
                    } else {
                        meshstructs[i].effectpoint = false;
                    }
                }
                if (needsfix) {
                    Fixup(obj);
                }
            } else {
                // This is the classic case
                for (i = 0; i < meshstructs.size(); i++) {
                    meshstructs[i].Name = obj->GetObjectName(i);
                    if (obj->GetObjectVertexCount(i) == 1) {
                        meshstructs[i].effectpoint = true;
                        meshstructs[i].effectpoint_vert = obj->GetObjectVertex(i, 0).position;
                    } else {
                        meshstructs[i].effectpoint = false;
                    }
                }
            }
        } else
            Fixup(obj);
    }

    delete obj;

    return true;
}

bool cSCNFile::Save() {
    unsigned long i;
    error = CSCNFILE_NO_ERROR;

    if (filename == "") {
        error = CSCNFILE_ERROR_FILENAME;
        return false;
    }

    FILE *f = fopen(filename.c_str(), "wb");
    if (!f) {
        error = CSCNFILE_ERROR_FILENAME;
        return false;
    }

    version = VERSION_CSCNFILE;

    unsigned long objlen = objfile.length()+1;
    fwrite(&objlen, sizeof(objlen), 1, f);
    fwrite(objfile.c_str(), objlen, 1, f);

    unsigned long meshcount = meshstructs.size();
    fwrite(&meshcount, sizeof(meshcount), 1, f);
    for (i = 0; i < meshcount; i++) {
        cMeshStruct ms = meshstructs[i];
        fwrite(&ms.disabled, sizeof(ms.disabled), 1, f);
        if (ms.disabled == false) {
            fwrite(&ms.flags, sizeof(ms.flags), 1, f);
            fwrite(&ms.place, sizeof(ms.place), 1, f);
            fwrite(&ms.unknown, sizeof(ms.unknown), 1, f);

            unsigned long ftxlen = ms.FTX.length()+1;
            fwrite(&ftxlen, sizeof(ftxlen), 1, f);
            if (ftxlen) {
                fwrite(ms.FTX.c_str(), ftxlen, 1, f);
            }

            unsigned long txslen = ms.TXS.length()+1;
            fwrite(&txslen, sizeof(txslen), 1, f);
            if (txslen) {
                fwrite(ms.TXS.c_str(), txslen, 1, f);
            }
        }
    }

    unsigned long texcount = flexitextures.size();
    fwrite(&texcount, sizeof(texcount), 1, f);
    for (i = 0; i < texcount; i++) {
        cFlexiTexture ft = flexitextures[i];
        fwrite(&ft.Recolorable, sizeof(ft.Recolorable), 1, f);

        unsigned long namelen = ft.Name.length()+1;
        fwrite(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            fwrite(ft.Name.c_str(), namelen, 1, f);
        }

        unsigned long texturelen = ft.Texture.length()+1;
        fwrite(&texturelen, sizeof(texturelen), 1, f);
        if (texturelen) {
            fwrite(ft.Texture.c_str(), texturelen, 1, f);
        }

        unsigned long alphalen = ft.Alpha.length()+1;
        fwrite(&alphalen, sizeof(alphalen), 1, f);
        if (alphalen) {
            fwrite(ft.Alpha.c_str(), alphalen, 1, f);
        }
    }

    unsigned long refcount = references.size();
    fwrite(&refcount, sizeof(refcount), 1, f);
    for (i = 0; i < refcount; i++) {
        unsigned long reflen = references[i].length()+1;
        fwrite(&reflen, sizeof(reflen), 1, f);
        if (reflen) {
            fwrite(references[i].c_str(), reflen, 1, f);
        }
    }

    // Old effectpoints store
    unsigned long ptcount = 0;
    fwrite(&ptcount, sizeof(ptcount), 1, f);

    fwrite(&version, sizeof(long), 1, f);

    // SCN Verison 1 Stuff
    long fixorient = 0;
    fwrite(&fixorient, sizeof(long), 1, f);

    // SCN Version 2 Stuff
    fwrite(&sivsettings.sivflags, sizeof(unsigned int), 1, f);
    fwrite(&sivsettings.sway, sizeof(float), 1, f);
    fwrite(&sivsettings.brightness, sizeof(float), 1, f);
    fwrite(&sivsettings.unknown, sizeof(float), 1, f);
    fwrite(&sivsettings.scale, sizeof(float), 1, f);

    // SCN Version 3 Stuff
    // Write Mesh names
    for (i = 0; i < meshcount; i++) {
        unsigned long namelen = meshstructs[i].Name.length()+1;
        fwrite(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            fwrite(meshstructs[i].Name.c_str(), namelen, 1, f);
        }
    }

    // Write transformations
    unsigned long matrices = transforms.size();
    fwrite(&matrices, sizeof(matrices), 1, f);
    for (unsigned long j = 0; j < matrices; j++) {
        unsigned long namelen = transformnames[j].length()+1;
        fwrite(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            fwrite(transformnames[j].c_str(), namelen, 1, f);
        }

        D3DMATRIX m = transforms[j];

        fwrite(&m._11, sizeof(m._11), 1, f);
        fwrite(&m._12, sizeof(m._12), 1, f);
        fwrite(&m._13, sizeof(m._13), 1, f);
        fwrite(&m._14, sizeof(m._14), 1, f);
        fwrite(&m._21, sizeof(m._21), 1, f);
        fwrite(&m._22, sizeof(m._22), 1, f);
        fwrite(&m._23, sizeof(m._23), 1, f);
        fwrite(&m._24, sizeof(m._24), 1, f);
        fwrite(&m._31, sizeof(m._31), 1, f);
        fwrite(&m._32, sizeof(m._32), 1, f);
        fwrite(&m._33, sizeof(m._33), 1, f);
        fwrite(&m._34, sizeof(m._34), 1, f);
        fwrite(&m._41, sizeof(m._41), 1, f);
        fwrite(&m._42, sizeof(m._42), 1, f);
        fwrite(&m._43, sizeof(m._43), 1, f);
        fwrite(&m._44, sizeof(m._44), 1, f);
    }

    // Write new effect point structures
    ptcount = effectpoints.size();
    fwrite(&ptcount, sizeof(ptcount), 1, f);
    for (i = 0; i < ptcount; i++) {
        cEffectPoint e = effectpoints[i];

        unsigned long namelen = e.Name.length()+1;
        fwrite(&namelen, sizeof(namelen), 1, f);
        if (namelen) {
            fwrite(e.Name.c_str(), namelen, 1, f);
        }

        matrices = e.Transform.size();
        fwrite(&matrices, sizeof(matrices), 1, f);
        for (unsigned long j = 0; j < matrices; j++) {
            D3DMATRIX m = e.Transform[j];

            namelen = e.TransformNames[j].length()+1;
            fwrite(&namelen, sizeof(namelen), 1, f);
            if (namelen) {
                fwrite(e.TransformNames[j].c_str(), namelen, 1, f);
            }

            fwrite(&m._11, sizeof(m._11), 1, f);
            fwrite(&m._12, sizeof(m._12), 1, f);
            fwrite(&m._13, sizeof(m._13), 1, f);
            fwrite(&m._14, sizeof(m._14), 1, f);
            fwrite(&m._21, sizeof(m._21), 1, f);
            fwrite(&m._22, sizeof(m._22), 1, f);
            fwrite(&m._23, sizeof(m._23), 1, f);
            fwrite(&m._24, sizeof(m._24), 1, f);
            fwrite(&m._31, sizeof(m._31), 1, f);
            fwrite(&m._32, sizeof(m._32), 1, f);
            fwrite(&m._33, sizeof(m._33), 1, f);
            fwrite(&m._34, sizeof(m._34), 1, f);
            fwrite(&m._41, sizeof(m._41), 1, f);
            fwrite(&m._42, sizeof(m._42), 1, f);
            fwrite(&m._43, sizeof(m._43), 1, f);
            fwrite(&m._44, sizeof(m._44), 1, f);
        }

    }

    fclose(f);
    return true;
}
