///////////////////////////////////////////////////////////////////////////////
//
// 3D Utility Library
// Copyright (C) 2006 Tobias Minch
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Written by
//   Tobias Minich - belgabor@gmx.de
//
///////////////////////////////////////////////////////////////////////////////

#include <vector>
#include "matrix.h"
#include <math.h>
#include <float.h>

float Deg2Rad(float deg) {
    return deg * M_PI / 180.0;
}

D3DVECTOR vectorInvert(const D3DVECTOR v) {
    D3DVECTOR r;
    r.x = -v.x;
    r.y = -v.y;
    r.z = -v.z;
    return r;
}

D3DMATRIX matrixGetUnity() {
    D3DMATRIX m;
    memset(&m, 0, sizeof(m));
    m._11 = 1.0;
    m._22 = 1.0;
    m._33 = 1.0;
    m._44 = 1.0;
    return m;
}

D3DMATRIX matrixGetLegacyFixOrientation() {
    D3DMATRIX m;
    memset(&m, 0, sizeof(m));
    m._13 = -1.0;
    m._21 = 1.0;
    m._32 = 1.0;
    m._44 = 1.0;
    return m;
}

D3DMATRIX matrixGetTranslation(float x, float y, float z) {
    D3DMATRIX m = matrixGetUnity();
    m._41 = x;
    m._42 = y;
    m._43 = z;
    return m;
}

D3DMATRIX matrixGetTranslation(D3DVECTOR v) {
    return matrixGetTranslation(v.x, v.y, v.z);
}

D3DMATRIX matrixGetScale(float x, float y, float z) {
    D3DMATRIX m = matrixGetUnity();
    m._11 = x;
    m._22 = y;
    m._33 = z;
    return m;
}

D3DMATRIX matrixGetScale(D3DVECTOR v) {
    return matrixGetScale(v.x, v.y, v.z);
}

D3DMATRIX matrixGetRotationX(float rad) {
    D3DMATRIX m = matrixGetUnity();
    m._22 = cos(rad);
    m._23 = -sin(rad);
    m._32 = sin(rad);
    m._33 = cos(rad);
    return m;
}
D3DMATRIX matrixGetRotationY(float rad) {
    D3DMATRIX m = matrixGetUnity();
    m._11 = cos(rad);
    m._31 = -sin(rad);
    m._13 = sin(rad);
    m._33 = cos(rad);
    return m;
}
D3DMATRIX matrixGetRotationZ(float rad) {
    D3DMATRIX m = matrixGetUnity();
    m._11 = cos(rad);
    m._12 = -sin(rad);
    m._21 = sin(rad);
    m._22 = cos(rad);
    return m;
}

float matrixCalcDeterminant(const D3DMATRIX *m) {
    return (m->_11*m->_22*m->_33*m->_44)
           -(m->_11*m->_22*m->_34*m->_43)
           +(m->_11*m->_23*m->_34*m->_42)
           -(m->_11*m->_23*m->_32*m->_44)
           +(m->_11*m->_24*m->_32*m->_43)
           -(m->_11*m->_24*m->_33*m->_42)
           -(m->_12*m->_23*m->_34*m->_41)
           +(m->_12*m->_23*m->_31*m->_44)
           -(m->_12*m->_24*m->_31*m->_43)
           +(m->_12*m->_24*m->_33*m->_41)
           -(m->_12*m->_21*m->_33*m->_44)
           +(m->_12*m->_21*m->_34*m->_43)
           +(m->_13*m->_24*m->_31*m->_42)
           -(m->_13*m->_24*m->_32*m->_41)
           +(m->_13*m->_21*m->_32*m->_44)
           -(m->_13*m->_21*m->_34*m->_42)
           +(m->_13*m->_22*m->_34*m->_41)
           -(m->_13*m->_22*m->_31*m->_44)
           -(m->_14*m->_21*m->_32*m->_43)
           +(m->_14*m->_21*m->_33*m->_42)
           -(m->_14*m->_22*m->_33*m->_41)
           +(m->_14*m->_22*m->_31*m->_43)
           -(m->_14*m->_23*m->_31*m->_42)
           +(m->_14*m->_23*m->_32*m->_41);
}

D3DMATRIX matrixMultiply(const D3DMATRIX *m1, const D3DMATRIX *m2) {
    D3DMATRIX m;
    m._11 = (m1->_11 * m2->_11) + (m1->_12 * m2->_21) + (m1->_13 * m2->_31) + (m1->_14 * m2->_41);
    m._12 = (m1->_11 * m2->_12) + (m1->_12 * m2->_22) + (m1->_13 * m2->_32) + (m1->_14 * m2->_42);
    m._13 = (m1->_11 * m2->_13) + (m1->_12 * m2->_23) + (m1->_13 * m2->_33) + (m1->_14 * m2->_43);
    m._14 = (m1->_11 * m2->_14) + (m1->_12 * m2->_24) + (m1->_13 * m2->_34) + (m1->_14 * m2->_44);

    m._21 = (m1->_21 * m2->_11) + (m1->_22 * m2->_21) + (m1->_23 * m2->_31) + (m1->_24 * m2->_41);
    m._22 = (m1->_21 * m2->_12) + (m1->_22 * m2->_22) + (m1->_23 * m2->_32) + (m1->_24 * m2->_42);
    m._23 = (m1->_21 * m2->_13) + (m1->_22 * m2->_23) + (m1->_23 * m2->_33) + (m1->_24 * m2->_43);
    m._24 = (m1->_21 * m2->_14) + (m1->_22 * m2->_24) + (m1->_23 * m2->_34) + (m1->_24 * m2->_44);

    m._31 = (m1->_31 * m2->_11) + (m1->_32 * m2->_21) + (m1->_33 * m2->_31) + (m1->_34 * m2->_41);
    m._32 = (m1->_31 * m2->_12) + (m1->_32 * m2->_22) + (m1->_33 * m2->_32) + (m1->_34 * m2->_42);
    m._33 = (m1->_31 * m2->_13) + (m1->_32 * m2->_23) + (m1->_33 * m2->_33) + (m1->_34 * m2->_43);
    m._34 = (m1->_31 * m2->_14) + (m1->_32 * m2->_24) + (m1->_33 * m2->_34) + (m1->_34 * m2->_44);

    m._41 = (m1->_41 * m2->_11) + (m1->_42 * m2->_21) + (m1->_43 * m2->_31) + (m1->_44 * m2->_41);
    m._42 = (m1->_41 * m2->_12) + (m1->_42 * m2->_22) + (m1->_43 * m2->_32) + (m1->_44 * m2->_42);
    m._43 = (m1->_41 * m2->_13) + (m1->_42 * m2->_23) + (m1->_43 * m2->_33) + (m1->_44 * m2->_43);
    m._44 = (m1->_41 * m2->_14) + (m1->_42 * m2->_24) + (m1->_43 * m2->_34) + (m1->_44 * m2->_44);
    return m;
}

D3DMATRIX *matrixMultiplyIP(D3DMATRIX *m1, const D3DMATRIX *m2) {
    D3DMATRIX m = matrixMultiply(m1, m2);
    *m1 = m;
    return m1;
}

D3DMATRIX matrixMultiply(D3DMATRIX *ms, unsigned int count) {
    D3DMATRIX m;
    switch (count) {
    case 0:
        m = matrixGetUnity();
        break;
    case 1:
        m = *ms;
        break;
    case 2:
        m = matrixMultiply(&ms[0], &ms[1]);
        break;
    default: {
        m = *ms;
        for (unsigned int i = 1; i < count; i++)
            matrixMultiplyIP(&m, &ms[i]);
    }
    break;
    }
    return m;
}

D3DMATRIX *matrixMultiplyIP(D3DMATRIX *ms, unsigned int count) {
    switch (count) {
    case 0:
    case 1:
        break;
    case 2:
        matrixMultiplyIP(&ms[0], &ms[1]);
        break;
    default: {
        for (unsigned int i = 1; i < count; i++)
            matrixMultiplyIP(ms, &ms[i]);
    }
    break;
    }
    return ms;
}

D3DMATRIX matrixMultiply(const std::vector<D3DMATRIX> ms) {
    D3DMATRIX m;
    switch (ms.size()) {
    case 0:
        m = matrixGetUnity();
        break;
    case 1:
        m = ms[0];
        break;
    case 2:
        m = matrixMultiply(&ms[0], &ms[1]);
        break;
    default: {
        m = ms[0];
        for (unsigned int i = 1; i < ms.size(); i++)
            matrixMultiplyIP(&m, &ms[i]);
    }
    break;
    }
    return m;
}


VERTEX matrixApply(VERTEX *v, const D3DMATRIX *m) {
    VERTEX r = *v;
    r.position.x = v->position.x * m->_11 + v->position.y * m->_21 + v->position.z * m->_31 + m->_41;
    r.position.y = v->position.x * m->_12 + v->position.y * m->_22 + v->position.z * m->_32 + m->_42;
    r.position.z = v->position.x * m->_13 + v->position.y * m->_23 + v->position.z * m->_33 + m->_43;
    r.normal.x = v->normal.x * m->_11 + v->normal.y * m->_21 + v->normal.z * m->_31 + m->_41;
    r.normal.y = v->normal.x * m->_12 + v->normal.y * m->_22 + v->normal.z * m->_32 + m->_42;
    r.normal.z = v->normal.x * m->_13 + v->normal.y * m->_23 + v->normal.z * m->_31 + m->_43;
    return r;
}

VERTEX *matrixApplyIP(VERTEX *v, const D3DMATRIX *m) {
    VERTEX r = matrixApply(v, m);
    *v = r;
    return v;
}

VERTEX2 matrixApply(VERTEX2 *v, const D3DMATRIX *m) {
    VERTEX2 r = *v;
    r.position.x = v->position.x * m->_11 + v->position.y * m->_21 + v->position.z * m->_31 + m->_41;
    r.position.y = v->position.x * m->_12 + v->position.y * m->_22 + v->position.z * m->_32 + m->_42;
    r.position.z = v->position.x * m->_13 + v->position.y * m->_23 + v->position.z * m->_33 + m->_43;
    r.normal.x = v->normal.x * m->_11 + v->normal.y * m->_21 + v->normal.z * m->_31 + m->_41;
    r.normal.y = v->normal.x * m->_12 + v->normal.y * m->_22 + v->normal.z * m->_32 + m->_42;
    r.normal.z = v->normal.x * m->_13 + v->normal.y * m->_23 + v->normal.z * m->_31 + m->_43;
    return r;
}

VERTEX2 *matrixApplyIP(VERTEX2 *v, const D3DMATRIX *m) {
    VERTEX2 r = matrixApply(v, m);
    *v = r;
    return v;
}

bool matrixIsEqual(const D3DMATRIX m1, const D3DMATRIX m2) {
    return (m1._11 == m2._11) && (m1._12 == m2._12) && (m1._13 == m2._13) && (m1._14 == m2._14) &&
           (m1._21 == m2._21) && (m1._22 == m2._22) && (m1._23 == m2._23) && (m1._24 == m2._24) &&
           (m1._31 == m2._31) && (m1._32 == m2._32) && (m1._33 == m2._33) && (m1._34 == m2._34) &&
           (m1._41 == m2._41) && (m1._42 == m2._42) && (m1._43 == m2._43) && (m1._44 == m2._44);
}


void boundsInit(D3DVECTOR *bbox_min, D3DVECTOR *bbox_max) {
    bbox_min->x = FLT_MAX;
    bbox_min->y = FLT_MAX;
    bbox_min->z = FLT_MAX;
    bbox_max->x = FLT_MIN;
    bbox_max->y = FLT_MIN;
    bbox_max->z = FLT_MIN;
}

void boundsContain(const D3DVECTOR *v, D3DVECTOR *bbox_min, D3DVECTOR *bbox_max) {
    if (v->x < bbox_min->x)
        bbox_min->x = v->x;
    if (v->y < bbox_min->y)
        bbox_min->y = v->y;
    if (v->z < bbox_min->z)
        bbox_min->z = v->z;
    if (v->x > bbox_max->x)
        bbox_max->x = v->x;
    if (v->y > bbox_max->y)
        bbox_max->y = v->y;
    if (v->z > bbox_max->z)
        bbox_max->z = v->z;
}

void boundsContain(const D3DVECTOR *inner_min, const D3DVECTOR *inner_max, D3DVECTOR *bbox_min, D3DVECTOR *bbox_max) {
    if (inner_min->x < bbox_min->x)
        bbox_min->x = inner_min->x;
    if (inner_min->y < bbox_min->y)
        bbox_min->y = inner_min->y;
    if (inner_min->z < bbox_min->z)
        bbox_min->z = inner_min->z;
    if (inner_max->x > bbox_max->x)
        bbox_max->x = inner_max->x;
    if (inner_max->y > bbox_max->y)
        bbox_max->y = inner_max->y;
    if (inner_max->z > bbox_max->z)
        bbox_max->z = inner_max->z;
}
